/*
 * Learn Alphabets
 *      The Learn Alphabets game
 *
 * Ken Shirriff         ken.shirriff@eng.sun.com
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * Please refer to the file "license.txt"
 * for further important copyright and licensing information.
 */

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "learn.h"
#define AppType    'KSla'
#define AppDBType   'Data'
#define AppVersion  1

#define assert(x) if (!(x)) abort("assert failed")

#if 1
#define abort(msg)      ErrDisplayFileLineMsg(__FILE__, __LINE__, msg)
#else
#define abort(msg)      ErrDisplayFileLineMsg(__FILE__, __LINE__, "")
#endif

#define disp(str) WinDrawChars(str, StrLen(str), 10, 10)

static void StartApplication(void);
static void StopApplication(void);
static void displayMoveCount(void);
static void EventLoop(void);
static Boolean learnEvent(EventPtr event);
static Boolean reviewEvent(EventPtr event);
static Boolean selectEvent(EventPtr event);
static void DrawBitmap(int resID, short x, short y);
static void *openDB(LocalID dbID);
static void findDBs();
static void closeDB(void);
static void updateStatus(void);
static void drawReview(int line);

struct datahdr {
  long version;
  long nbitmaps;
  long x;
  long y;
  long c;
};

int num_right = 0;
int num_wrong = 0;

int currentItem = -1;

struct state {
  unsigned int version;
};

struct state board = {
  AppVersion
};

// Main loop
DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  if (cmd == sysAppLaunchCmdNormalLaunch)
 {
    StartApplication();

    FrmGotoForm(formID_select);

    EventLoop();  // Event loop

    StopApplication();
  }
  return 0;
}

struct bitmaphdr {
	short x;
	short y;
	short c;
	short pad[5];
};

struct entry {
	int qoffset;
	int aoffset;
};

// Maximum number of database files
#define MAXDB 32

#define QX 0		// X position for the question
#define QY 20		// Y position for the question
#define AY 40		// Y position for the answer
#define A_INDENT 25	// Indent for the answer
#define AH 20		// Line spacing
#define STR_OFF 5	// Offset for the string
#define NUM_A 5

char *items[MAXDB];
int nitems = 0;
int dir = 0;

DmOpenRef dbR = NULL;
VoidHand hand;
struct datahdr *hdr;

static void *openDB(LocalID dbID) {
  closeDB();
  dbR = DmOpenDatabase(0, dbID, dmModeReadOnly);
  assert(dbR != 0);
  assert (DmNumRecords(dbR)>=1);
  hand = DmGetRecord(dbR, 0);
  return MemHandleLock(hand);
}

static void closeDB(void) {
  if (dbR != NULL) {
    MemHandleUnlock(hand);
    DmReleaseRecord(dbR, 0, 0);
    DmCloseDatabase(dbR);
    dbR = NULL;
    hand = NULL;
  }
}

int numstr;
char **strings = NULL;
short *permutation;
int qoffset = 0;

static void getData(int idx)
{
  VoidHand hand;
  int i;
  char *p;
  LocalID dbID = DmFindDatabase(0, items[idx]);
  assert(dbID != 0);
  hdr = (struct datahdr *)openDB(dbID);
  assert(hdr->version == 1234);
  // Count the strings
  numstr = hdr->nbitmaps;
  // Set up the string pointers
  if (strings != NULL) {
    MemPtrFree(strings);
    MemPtrFree(permutation);
  }
  strings = (char **)MemPtrNew(sizeof(char *)*numstr);
  permutation = MemPtrNew(sizeof(short *)*numstr);
  p = ((char *)hdr)+ sizeof(struct datahdr)+ hdr->nbitmaps*hdr->y*hdr->c;
  for (i=0; i<numstr ; i++) {
    strings[i] = p;
    p += StrLen(p)+1;
  }
  if (numstr<NUM_A) {
    abort("not enough strings");
  } else if (strings[0]==0) {
    abort("null string pointer");
  } else if (strings[0][0]==0) {
    abort("null string");
  } else if (numstr != hdr->nbitmaps) {
    abort("wrong number of strings");
  }
}

static void clear(int x, int y, int w, int h) {
  RectangleType r;
  r.topLeft.x = x;
  r.topLeft.y = y;
  r.extent.x = w;
  r.extent.y = h;
  WinEraseRectangle(&r, 0);
}

int right_choice;
int answers[NUM_A];
int show[NUM_A];
static void pickQuestion() {
  int i,j;
  int t;
  clear(0, 22, 160, 160);
  answers[0] = permutation[qoffset];
  if (answers[0]>=numstr) {
    abort("permutation");
  }
  show[0] = 0;

  // Select 4 different wrong answers
  for (i=1; i<NUM_A;i++) {
    answers[i] = SysRandom(0)%numstr;
    if (answers[i]>=numstr) {
      abort("a");
    }
    show[i] = 0;
    for (j=0; j<i; j++) {
      if (answers[j] == answers[i]) {
	i--;
	break;
      }
    }
  }

  // Switch the right answer with a random slot
  right_choice = SysRandom(0)%NUM_A;
  t = answers[right_choice];
  if (t>=numstr) {
    abort("c");
  }
  answers[right_choice] = answers[0];
  answers[0] = t;
  // Advance the pointer to the next question
  qoffset++;
}

char *status = "";
int status_bitmap;
char *status_str;

static void drawQuiz() {
  int i;

  updateStatus();
  clear(0,140,160,20);
  WinDrawLine(0,140,160,140);
  WinDrawChars(status, StrLen(status), 5, 142);
  if (status_bitmap != -1) {
    if (status_bitmap<0 || status_bitmap>= hdr->nbitmaps) {
      abort("badID");
    }
    DrawBitmap(status_bitmap, QX, 141);
    WinDrawChars(status_str, StrLen(status_str),QX+A_INDENT, 142);
  }

  if (dir==0) {
    if (answers[right_choice]<0 || answers[right_choice]>= hdr->nbitmaps) {
      abort("badID");
    }
    DrawBitmap(answers[right_choice], QX, QY);
  } else {
      WinDrawChars(strings[answers[right_choice]], StrLen(strings[answers[right_choice]]), QX, QY+STR_OFF);
  }
  for (i=0;i<NUM_A;i++) {
    if (dir==1 || show[i]) {
      DrawBitmap(answers[i], QX, AY+AH*i);
    }
    if (dir==0 || show[i]) {
      if (strings[answers[i]]==NULL) {
        abort("null");
      }
      WinDrawChars(strings[answers[i]], StrLen(strings[answers[i]]), QX+A_INDENT, AY+AH*i+STR_OFF);
    }
  }
}

#define NUM_REVIEW 7

// Draw review characters
static void drawReview(int line) {
  int i;

  clear(0,15,160,160);
  for (i=0; i<NUM_REVIEW && i+line<hdr->nbitmaps; i++) {
    DrawBitmap(i+line, QX, QY+AH*i);
    WinDrawChars(strings[i+line], StrLen(strings[i+line]), QX+A_INDENT,
      QY+AH*i+STR_OFF);
  }
}

static void permute() {
  int i,j;
  int t;
  for (i=0;i<numstr;i++) {
    permutation[i] = i;
  }
  for (i=0;i<numstr-1;i++) {
    j = i+(SysRandom(0)%(numstr-i));
    t = permutation[i];
    permutation[i] = permutation[j];
    permutation[j] = t;
  }
}

void restart(void) {
  int i;
  qoffset = 0;
  num_right = 0;
  num_wrong = 0;
  if (hdr->nbitmaps != numstr) {
    abort("???");
  }
  permute();
  if (hdr->nbitmaps != numstr) {
    abort("???");
  }
  for (i=0;i<numstr;i++) {
    if (permutation[i]>=numstr) {
      abort("permutation");
    }
  }
  status="";
  status_bitmap = -1;
  status_str = "";
  pickQuestion();
  if (hdr->nbitmaps != numstr) {
    abort("???");
  }
  if (answers[right_choice] >= hdr->nbitmaps) {
    abort("answers?");
  }
  for (i=0;i<NUM_A;i++) {
    if (answers[i] >= hdr->nbitmaps) {
      abort("answers?");
    }
  }
  if (status_bitmap != -1) {
    abort("bitmap?");
  }
  {
  for (i=0;i<NUM_A;i++) {
    if (answers[i] >= hdr->nbitmaps) {
      abort("answers");
    }
  }
  }
  drawQuiz();
}

static void DrawBitmap(int n, short x, short y)
{
  BitmapPtr resP;
  if (n<0 || n>= hdr->nbitmaps) {
    abort("badID");
    return;
  }
  resP = MemPtrNew(sizeof(BitmapType)+hdr->y*hdr->c);
  resP->width = hdr->x;
  resP->height = hdr->y;
  resP->rowBytes = hdr->c;
  resP->flags.compressed = 0;
  resP->flags.reserved = 0;
  resP->reserved[0] = 0;
  resP->reserved[1] = 0;
  resP->reserved[2] = 0;
  resP->reserved[3] = 0;
  MemMove(resP+1, ((char *)hdr)+sizeof(struct datahdr)+n*hdr->y*hdr->c,
      hdr->y*hdr->c);
  WinDrawBitmap(resP, x, y);
  MemPtrFree(resP);
}

#define statusLeft 90
#define statusTop 16
#define statusWidth 70
#define statusHeight 11

// Display move count
static void updateStatus(void) {
  char buf[40];
  RectangleType r;
  r.topLeft.x = statusLeft;
  r.topLeft.y = statusTop;
  r.extent.x = statusWidth;
  r.extent.y = statusHeight;
  WinEraseRectangle(&r, 0);
  StrCopy(buf, "Score: ");
  StrIToA(buf+StrLen(buf), num_right);
  StrCat(buf, " of ");
  StrIToA(buf+StrLen(buf), num_right+num_wrong);
  WinDrawChars(buf, StrLen(buf), statusLeft, statusTop);
}

// User got question wrong, so put it back on the list.
static void putbackQuestion() {
  // qoffset-1 is the question that was wrong
  int pos, t;
  qoffset -= 1;
  assert(qoffset>=0 && qoffset<numstr);
  pos = qoffset+SysRandom(0)%(numstr-qoffset);
  t = permutation[qoffset];
  permutation[qoffset] = permutation[pos];
  permutation[pos] = t;
}

// Handle a click on an answer
static int PenEvent(enum events etype, short nx, short ny)
{
  short i;
  if (etype == penDownEvent) {
    if (ny>AY) {
      int n = (ny-AY)/AH;
      if (n<NUM_A) {
	if (n==right_choice) {
	  if (status_bitmap == -1) {
	    putbackQuestion();
	  }
	  status = "";
	  status_bitmap = answers[right_choice];
	  status_str = strings[answers[right_choice]];
	  num_right++;
	  if (qoffset==numstr) {
	    FrmAlert(alertID_done);
	    restart();
	  }
	  pickQuestion();
	} else {
	  status = "Sorry";
	  show[n] = 1;
	  status_bitmap = -1;
	  status_str = "";
	  num_wrong++;
	}
	drawQuiz();
	return true;
      }
    }
  }
  return false;
}

struct pref_t {
	long foo;
	long bar;
} prefs;

static void StartApplication(void) {
  if (!PrefGetAppPreferencesV10(AppType, AppVersion, &prefs, sizeof(prefs))) {
  }
  prefs.foo = 0x11223344;
  prefs.bar = 0x55667788;
}

static void StopApplication(void) {
  closeDB();
  PrefSetAppPreferencesV10(AppType, AppVersion, &prefs, sizeof(prefs));
}


// Look up the appropriate databases
static void findDBs()
{
  DmSearchStateType state;
  Boolean newSearch = 1;
  UInt idx;
  UInt cardNo;
  LocalID dbID;
  Err err;
  DmOpenRef ref;
  VoidHand hand;
  char *p;
  int i;
  LocalID id;
  ListPtr listp;
  VoidHand resH;
  FormPtr frm;
  char name[32];

  for (i=0;i<nitems;i++) {
    MemPtrFree(items[i]);
  }
  nitems = 0;

  for (i=0;;i++) {
    char *ptr;
    err = DmGetNextDatabaseByTypeCreator(newSearch, &state, AppDBType,
      AppType, 0, &cardNo, &dbID);
    if (err == dmErrCantFind) {
      break;
    }
    newSearch = 0;
    err = DmDatabaseInfo(cardNo, dbID, name, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL);
    ptr = MemPtrNew(StrLen(name)+1);
    MemMove(ptr, name, StrLen(name)+1);
    items[nitems++] = ptr;
  }
  frm = FrmGetActiveForm();
  listp = (ListPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_data));
  if (listp == 0) {
    abort("No listp");
  }
  LstSetListChoices(listp, items, nitems);
  LstSetHeight(listp, nitems>8?8:nitems);
}

static void EventLoop(void)
{
  short err;
  int formID;
  FormPtr form;
  EventType event;

  do
  {

    EvtGetEvent(&event, evtWaitForever);

    if (SysHandleEvent(&event)) continue;
    if (MenuHandleEvent((void *)0, &event, &err)) continue;

    if (event.eType == frmLoadEvent)
    {
      formID = event.data.frmLoad.formID;
      form = FrmInitForm(formID);
      FrmSetActiveForm(form);
      switch (formID) 
      {
      case formID_review:
        FrmSetEventHandler(form, (FormEventHandlerPtr) reviewEvent);
        break;
      case formID_learn:
        FrmSetEventHandler(form, (FormEventHandlerPtr) learnEvent);
        break;
      case formID_select:
        FrmSetEventHandler(form, (FormEventHandlerPtr) selectEvent);
        break;
      }
    }

    FrmDispatchEvent(&event);
  } while(event.eType != appStopEvent);
}

static Boolean learnEvent(EventPtr event)
{
  FormPtr   form;
  int       handled = 0;

  switch (event->eType) 
  {
  case frmOpenEvent:
    form = FrmGetActiveForm();
    FrmDrawForm(form);
    getData(currentItem);
    restart();
    handled = 1;
    break;
    
  case menuEvent:
    switch (event->data.menu.itemID)
    {
    case menuitemID_about:
      FrmAlert(alertID_about);
      break;
    case menuitemID_instr:
      FrmAlert(alertID_instr);
      break;
    case menuitemID_flip:
      dir = 1-dir;
      clear(0, 15, 160, 160);
      drawQuiz();
      break;
    case menuitemID_restart:
      closeDB();
      restart();
      break;
    case menuitemID_select:
      FrmGotoForm(formID_select);
      break;
    }
    handled = 1;
    break;

  case lstSelectEvent:
    FrmAlert(alertID_instr);
    handled = 1;
    break;
  case penDownEvent:
    handled = PenEvent(event->eType, event->screenX, event->screenY);
    break;

  }
  return handled;
}

// Helper function
static void *getObjectPtr(Word objID) {
  FormPtr frm = FrmGetActiveForm();
  return FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, objID));
}

int reviewline = 0;

// Set up/down buttons correctly
static Boolean reviewButtons() {
  if (reviewline+NUM_REVIEW < hdr->nbitmaps) {
    CtlSetLabel(getObjectPtr(buttonID_down), "\002");
  } else {
    CtlSetLabel(getObjectPtr(buttonID_down), "\004");
  }
  if (reviewline>=NUM_REVIEW) {
    CtlSetLabel(getObjectPtr(buttonID_up), "\001");
  } else {
    CtlSetLabel(getObjectPtr(buttonID_up), "\003");
  }
  CtlDrawControl(getObjectPtr(buttonID_up));
  CtlDrawControl(getObjectPtr(buttonID_down));
}

// Review all the data
static Boolean reviewEvent(EventPtr event)
{
  FormPtr   form;
  int       handled = 0;

  switch (event->eType) 
  {
  case frmOpenEvent:
    form = FrmGetActiveForm();
    FrmDrawForm(form);
    getData(currentItem);
    reviewline = 0;
    handled = 1;
    drawReview(reviewline);
    reviewButtons();
    break;
    
  case menuEvent:
    switch (event->data.menu.itemID)
    {
    case menuitemID_about:
      FrmAlert(alertID_about);
      break;
    case menuitemID_select:
      FrmGotoForm(formID_select);
      break;
    }
    handled = 1;
    break;

  case ctlSelectEvent:
    switch (event->data.ctlSelect.controlID) {
      case buttonID_up:
	if (reviewline>=NUM_REVIEW) {
	  reviewline -= NUM_REVIEW;
	}
        drawReview(reviewline);
        reviewButtons();
	handled = 1;
	break;
      case buttonID_down:
	if (reviewline+NUM_REVIEW < hdr->nbitmaps) {
	  reviewline += NUM_REVIEW;
	}
        drawReview(reviewline);
        reviewButtons();
	handled = 1;
	break;
    }
  }
  return handled;
}

static void selectFormDraw(void) {
  findDBs();
  FrmDrawForm(FrmGetActiveForm());
}

static void deleteDB(int idx) {
  LocalID dbID;
  Err err;
  dbID = DmFindDatabase(0, items[idx]);
  err = DmDeleteDatabase(0, dbID);
  if (err != 0) {
    abort("delete Failed");
  }
  FrmGotoForm(formID_select);
}

static Boolean selectEvent(EventPtr event)
{
  switch (event->eType) 
  {
  case frmOpenEvent:
  case frmUpdateEvent:
    selectFormDraw();
    return true;
    break;
    
  case menuEvent:
    switch (event->data.menu.itemID)
    {
    case menuitemID_about:
      FrmAlert(alertID_about);
      return true;
    case menuitemID_instr:
      FrmAlert(alertID_instr);
      return true;
    }
    break;
  case ctlSelectEvent:
    {
      ListPtr listp;
      FormPtr frm = FrmGetActiveForm();
      listp = (ListPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_data));
      currentItem = LstGetSelection(listp);
      if (currentItem == -1) {
	return false;
      }
    switch (event->data.ctlSelect.controlID) {
      case buttonID_study:
	FrmGotoForm(formID_learn);
	return true;
      case buttonID_review:
	FrmGotoForm(formID_review);
	return true;
      case buttonID_delete:
	if (FrmAlert(alertID_delete)==1) {
	} else {
	  deleteDB(currentItem);
	}
	return true;
    }
  }
  break;

  }
  

  return false;
}
